import os
import subprocess
from PySide import QtGui, QtCore

import rospy
from rosbag import Bag
from rosgraph_msgs.msg import Clock


PLAYING = 'playing'
PAUSED = 'paused'
STOPPED = 'stopped'



def format_time(time):
    secs = time % 60
    mins = int(time / 60)
    return '%d:%04.1f' % (mins, secs)


class MediaButton(QtGui.QPushButton):
    def __init__(self, *args):
        super(MediaButton, self).__init__(*args)
        icon_size = QtCore.QSize(36, 28)
        self.setIconSize(icon_size)
        self.setStyleSheet("QPushButton { padding: 8px }")
    
    def set_icon_play(self):
        self.setIcon(QtGui.QIcon(self.style().standardIcon(QtGui.QStyle.SP_MediaPlay).pixmap(64)))
        self.setToolTip('Play')
    
    def set_icon_pause(self):
        self.setIcon(QtGui.QIcon(self.style().standardIcon(QtGui.QStyle.SP_MediaPause).pixmap(64)))
        self.setToolTip('Pause')
        
    def set_icon_stop(self):
        self.setIcon(QtGui.QIcon(self.style().standardIcon(QtGui.QStyle.SP_MediaStop).pixmap(64)))
        self.setToolTip('Stop')
        
    def set_icon_beginning(self):
        self.setIcon(QtGui.QIcon(self.style().standardIcon(QtGui.QStyle.SP_MediaSkipBackward).pixmap(64)))
        self.setToolTip('Go to the beginning')
        
    def set_icon_step(self):
        self.setIcon(QtGui.QIcon(self.style().standardIcon(QtGui.QStyle.SP_ArrowRight).pixmap(64)))
        self.setToolTip('Step forward')
        


class BagPlayer(object):
    def __init__(self, gui):
        self.gui = gui
        self.filename = gui.pkg_path + '/bag/ur_test_everything.bag' #default
        self.rosbag_proc = None
        self.is_looping = False
        self.state = STOPPED
        self.bag = None
        self.progress = 0.0
        
        rospy.Subscriber('/clock', Clock, self.clock_cb, queue_size=1)
        
        # @todo: Add bag information to GUI?
        
        
        ### Create the "playback options" GUI box
        self.options_box = QtGui.QGroupBox('Options')
        self.options_box.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed,
                                                         QtGui.QSizePolicy.Fixed))
        options_layout = QtGui.QVBoxLayout()
        
        file_box = QtGui.QGroupBox()
        file_box.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed,
                                                 QtGui.QSizePolicy.Fixed))
        file_layout = QtGui.QHBoxLayout()
        file_layout.setContentsMargins(0, 0, 0, 0)
        file_label = QtGui.QLabel('Select file:')
        self.file_edit = QtGui.QLineEdit(self.filename)
        self.file_edit.setFixedSize(400, 26)
        self.file_edit.setReadOnly(True)
        file_btn = QtGui.QPushButton('...')
        file_btn.setMaximumWidth(30)
        file_btn.clicked.connect(self.select_bag)
        file_layout.addWidget(file_label)
        file_layout.addWidget(self.file_edit)
        file_layout.addWidget(file_btn)
        file_box.setLayout(file_layout)
        file_box.setContentsMargins(5, 0, 0, 0)
        
        loop_box = QtGui.QGroupBox()
        loop_box.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed,
                                                 QtGui.QSizePolicy.Fixed))
        loop_layout = QtGui.QHBoxLayout()
        loop_layout.setContentsMargins(0, 0, 0, 0)
        self.loop_cb = QtGui.QCheckBox()
        self.loop_cb.setChecked(True)
        loop_label = QtGui.QLabel('Loop playback')
        loop_layout.addWidget(self.loop_cb)
        loop_layout.addWidget(loop_label)
        loop_box.setLayout(loop_layout)
        loop_box.setContentsMargins(0, 0, 0, 0)
        
        rate_box = QtGui.QGroupBox()
        rate_box.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed,
                                                 QtGui.QSizePolicy.Fixed))
        rate_layout = QtGui.QHBoxLayout()
        rate_layout.setContentsMargins(0, 0, 0, 0)
        rate_label = QtGui.QLabel('Playback speed (rate factor):')
        self.rate_edit = QtGui.QLineEdit('1')
        self.rate_edit.setFixedSize(34, 26)
        rate_layout.addWidget(rate_label)
        rate_layout.addWidget(self.rate_edit)
        rate_box.setLayout(rate_layout)
        rate_box.setContentsMargins(5, 0, 0, 0)
        
        options_layout.addWidget(file_box)
        options_layout.addWidget(loop_box)
        options_layout.addWidget(rate_box)
        self.options_box.setLayout(options_layout)
        
        
        
        ### Create the media box
        self.media_box = QtGui.QGroupBox('Player')
        media_layout = QtGui.QVBoxLayout()
        
        # Create the progress bar
        progress_box = QtGui.QGroupBox()
        progress_layout = QtGui.QHBoxLayout()
        self.progress_label = QtGui.QLabel()
        self.progress_bar = QtGui.QProgressBar()
        self.progress_bar.setFormat('')
        self.duration_label = QtGui.QLabel()
        progress_layout.addWidget(self.progress_label)
        progress_layout.addWidget(self.progress_bar)
        progress_layout.addWidget(self.duration_label)
        progress_box.setLayout(progress_layout)
        
        # Create the buttons
        buttons_box = QtGui.QGroupBox()
        buttons_layout = QtGui.QHBoxLayout()
        buttons_layout.setAlignment(QtCore.Qt.AlignHCenter)
        self.media_reset_btn = MediaButton()
        self.media_reset_btn.set_icon_beginning()
        self.media_reset_btn.clicked.connect(self.media_reset)
        self.media_playpause_btn = MediaButton()
        self.media_playpause_btn.set_icon_play()
        self.media_playpause_btn.clicked.connect(self.media_playpause)
        self.media_step_btn = MediaButton()
        self.media_step_btn.set_icon_step()
        self.media_step_btn.clicked.connect(self.media_step)
        self.media_stop_btn = MediaButton()
        self.media_stop_btn.set_icon_stop()
        self.media_stop_btn.clicked.connect(self.media_stop)
        buttons_layout.addWidget(self.media_reset_btn)
        buttons_layout.addWidget(self.media_playpause_btn)
        buttons_layout.addWidget(self.media_stop_btn)
        buttons_layout.addWidget(self.media_step_btn)
        buttons_layout.setSpacing(30)
        buttons_box.setLayout(buttons_layout)
        
        media_layout.addWidget(progress_box)
        media_layout.addWidget(buttons_box)
        self.media_box.setLayout(media_layout)
        
        
    
    
    def init(self):
        self.bag = Bag(self.filename, 'r')
        self.get_info()
        self.progress = self.bag._chunks[0].start_time.to_sec()
        if self.progress_bar is not None:
            self.progress_bar.setRange(0, 10*self.get_duration())
    
    
    def select_bag(self):
        if self.state is not STOPPED:
            # @todo: give error message dialog or something
            return
        
        name, _filter = QtGui.QFileDialog.getOpenFileName(self.gui,
                                            self.gui.tr('Choose bag file to open'),
                                            self.gui.pkg_path + '/bag/',
                                            self.gui.tr('Bag files (*.bag)'))
        
        if len(name) > 1:
            self.filename = name
            self.file_edit.setText(self.filename)
            self.init()
    
    def get_info(self):
        print "="*80
        print self.bag
        print "="*80
        
    def get_start_time(self):
        return self.bag._chunks[0].start_time.to_sec()
    
    def get_end_time(self):
        return self.bag._chunks[-1].end_time.to_sec()
    
    def get_duration(self):
        return round(self.get_end_time() - self.get_start_time(), 1)
    
    def get_progress(self):
        return round(self.progress - self.get_start_time(), 1)
    
    def clock_cb(self, msg):
        if self.state is PLAYING:
            self.progress = msg.clock.to_sec()

    def start(self):
        
        if not os.path.isfile(str(self.file_edit.text())):
            msg_box = QtGui.QMessageBox()
            msg_box.setText("No such bag file!")
            msg_box.exec_()
            return
        
        # @todo: check if topics are available ?
        
        ### Rosbag stuff
        args = ['rosbag', 'play', '--clock']
        if self.loop_cb.isChecked():
            args.append('--loop')
            self.is_looping = True
        else:
            self.is_looping = False
        args.append('--rate')
        args.append(str(self.rate_edit.text()))
        args.append(str(self.file_edit.text()))
        self.rosbag_proc = subprocess.Popen(args,
                                            stdin=subprocess.PIPE,
                                            stdout=open('/dev/null', 'w'),
                                            stderr=subprocess.STDOUT)
        self.state = PLAYING
        self.media_playpause_btn.set_icon_pause()
        self.options_box.setDisabled(True)
        
        # Set the 'playback_rate' parameter, to tell robot_proc
        rospy.set_param('/process_visualizer/playback_rate', str(self.rate_edit.text()))
    
    
    def stop(self):
        if self.state == STOPPED:
            return
        
        self.rosbag_proc.send_signal(subprocess.signal.SIGINT)
        self.progress = self.get_start_time()
        self.update_progress()
        self.state = STOPPED
        self.media_playpause_btn.set_icon_play()
        self.options_box.setEnabled(True)
    
    
    def media_reset(self):
        if self.state is PLAYING or self.state is PAUSED:
            self.stop()
        self.start()
    
    
    
    def media_playpause(self):
        if self.state is PLAYING:
            #pause
            self.rosbag_proc.stdin.write(' ')
            self.state = PAUSED
            self.media_playpause_btn.set_icon_play()
        
        elif self.state is PAUSED:
            #resume
            self.rosbag_proc.stdin.write(' ')
            self.state = PLAYING
            self.media_playpause_btn.set_icon_pause()
        
        elif self.state is STOPPED:
            #start
            self.start()
            

       
    def media_stop(self):
        self.stop()



    def media_step(self):
        if self.state is PAUSED:
            self.rosbag_proc.stdin.write('s')
    
    
    
    def update_progress(self):
        dur = self.get_duration()
        progress = self.get_progress()
        
        if progress == dur and not self.is_looping:
            self.stop()
            return
        if progress > dur:
            progress = dur
        elif progress < 0.0:
            progress = 0.0
        
        self.progress_label.setText(format_time(progress))
        self.progress_bar.setValue(10*progress)
        self.duration_label.setText(format_time(dur))


    def set_visible(self, visible=True):
        self.options_box.setVisible(visible)
        self.media_box.setVisible(visible)


