import subprocess
from PySide import QtGui, QtCore
from time import strftime

import rospy

from bagplayer import format_time

RECORDING = 'recording'
STOPPED = 'stopped'



class BagRecorder(object):
    def __init__(self, pkg_path):
        self.path = pkg_path + '/bag/'
        self.filename = ''
        self.rosbag_proc = None
        self.state = STOPPED
        self.topics = []
        self.starttime = rospy.Time()
        
        ### Create the "real-time options" GUI box
        self.options_box = QtGui.QGroupBox('Recording Options')
        self.options_box.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed,
                                                         QtGui.QSizePolicy.Fixed))
        options_layout = QtGui.QGridLayout()
        
        self.standard_radio = QtGui.QRadioButton(self.options_box)
        self.standard_radio.setChecked(True)
        self.prefix_radio = QtGui.QRadioButton(self.options_box)
        self.fullname_radio = QtGui.QRadioButton(self.options_box)
        
        standard_label = QtGui.QLabel('Use standard file name (current date and time)')
        
        prefix_box = QtGui.QGroupBox()
        prefix_box.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed,
                                                   QtGui.QSizePolicy.Fixed))
        prefix_layout = QtGui.QHBoxLayout()
        prefix_layout.setContentsMargins(0, 0, 0, 0)
        prefix_label = QtGui.QLabel('Use prefix:')
        self.prefix_edit = QtGui.QLineEdit('')
        self.prefix_edit.setFixedSize(200, 26)
        prefix_label2 = QtGui.QLabel('(+ current date and time)')
        prefix_layout.addWidget(prefix_label)
        prefix_layout.addWidget(self.prefix_edit)
        prefix_layout.addWidget(prefix_label2)
        prefix_box.setLayout(prefix_layout)
        prefix_box.setContentsMargins(0, 0, 0, 0)
        
        fullname_box = QtGui.QGroupBox()
        fullname_box.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed,
                                                     QtGui.QSizePolicy.Fixed))
        fullname_layout = QtGui.QHBoxLayout()
        fullname_layout.setContentsMargins(0, 0, 0, 0)
        fullname_label = QtGui.QLabel('Use following file name:')
        self.fullname_edit = QtGui.QLineEdit('')
        self.fullname_edit.setFixedSize(200, 26)
        fullname_layout.addWidget(fullname_label)
        fullname_layout.addWidget(self.fullname_edit)
        fullname_box.setLayout(fullname_layout)
        fullname_box.setContentsMargins(0, 0, 0, 0)
        
        options_layout.addWidget(self.standard_radio, 0, 0)
        options_layout.addWidget(standard_label, 0, 1)
        options_layout.addWidget(self.prefix_radio, 1, 0)
        options_layout.addWidget(prefix_box, 1, 1)
        options_layout.addWidget(self.fullname_radio, 2, 0)
        options_layout.addWidget(fullname_box, 2, 1)
        self.options_box.setLayout(options_layout)
        
        
        
        ### Create the start/stop record button
        # @todo: make progress label nicer....
        self.media_box = QtGui.QGroupBox()
        media_layout = QtGui.QVBoxLayout()
        self.progress_label = QtGui.QLabel('')
        self.record_btn = QtGui.QPushButton('Start recording')
        self.record_btn.clicked.connect(self.media_record)
        media_layout.addWidget(self.progress_label, 0, QtCore.Qt.AlignHCenter)
        media_layout.addWidget(self.record_btn, 0, QtCore.Qt.AlignHCenter)
        self.media_box.setLayout(media_layout)
        
        
        
    def init(self, topics=[]):
        self.topics = topics
    


    def start(self):
        if len(self.topics) > 0:
            
            # Use correct file name
            if self.standard_radio.isChecked():
                self.filename = self.path + strftime('%Y-%m-%d-%H-%M-%S')
            elif self.prefix_radio.isChecked():
                self.filename = ( self.path + str(self.prefix_edit.text()) + '_' +
                                  strftime('%Y-%m-%d-%H-%M-%S') )
            else:
                self.filename = self.path + str(self.fullname_edit.text())
            
            # Start rosbag
            args = ['rosbag', 'record', '-O']
            args.append(self.filename)
            args.extend(self.topics)
            self.rosbag_proc = subprocess.Popen(args,
                                                stdin=subprocess.PIPE)
                                                #stdout=open('/dev/null', 'w'),
                                                #stderr=subprocess.STDOUT)
            self.starttime = rospy.Time.now().to_sec()
            self.state = RECORDING
            self.options_box.setDisabled(True)
    
    
    def stop(self):
        if self.state is STOPPED:
            return
        
        self.rosbag_proc.send_signal(subprocess.signal.SIGINT)
        
        self.state = STOPPED
        self.options_box.setEnabled(True)


    def media_record(self):
        if self.state is RECORDING:
            #STOP recording
            self.stop()

            msg_box = QtGui.QMessageBox()
            msg_box.setText('Finished recording to ' + self.filename + '.bag')
            msg_box.exec_()
            self.record_btn.setText('Start recording')
        else:
            #START recording
            self.start()
            self.record_btn.setText('Stop recording')

    
    def update_progress(self):
        if self.state is RECORDING:
            self.progress_label.setText('Time: ' + format_time(
                                        rospy.Time.now().to_sec() - self.starttime) )
        


    def set_visible(self, visible=True):
        self.options_box.setVisible(visible)
        self.media_box.setVisible(visible)



