#!/usr/bin/python
# -*- coding: utf-8 -*-
import sys
import os
import signal
import time
import subprocess
from PySide import QtGui, QtCore

import roslib; roslib.load_manifest('process_visualizer')
import roslaunch
import rospy

import bagrecorder
import bagplayer
from xmlparser import XmlParser
from rvizconfig import RVizConfig


DEFAULT = 'default'
REALTIME = 'realtime'
PLAYBACK = 'playback'


def handle_sigint(*args):
    sys.stderr.write("\rSIGINT")
    QtGui.QApplication.quit()


def get_all_topics_to_record(devices):
    topics = []
    for device in devices:
        if device.gui_cb.isChecked():
            for topic in device.topics:
                if topic[2] == True:
                    topics.append(topic[0])
    return topics




class LauncherApp(QtGui.QDialog):
    def __init__(self, config_path):
        super(LauncherApp, self).__init__()
        
        ### Initialize variables
        self.state = DEFAULT
        self.pkg_path = roslib.packages.get_pkg_dir('process_visualizer')
        self.config_path = config_path
        self.recorder = bagrecorder.BagRecorder(self.pkg_path)
        self.player = bagplayer.BagPlayer(self)
        self.launcher = roslaunch.ROSLaunch()
        self.subprocesses = []
        
        self.devices = XmlParser(self.config_path).get_all()
        
        ### Create "devices" box
        self.devices_box = QtGui.QGroupBox('Devices')
        self.devices_box.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed,
                                                         QtGui.QSizePolicy.Fixed))
        devices_grid = QtGui.QGridLayout()
        self.devices_box.setLayout(devices_grid)
        
        for i in range(len(self.devices)):
            label = QtGui.QLabel(self.devices[i].name + ' (' + self.devices[i].type + ')')
            self.devices[i].gui_cb = QtGui.QCheckBox()
            self.devices[i].gui_cb.setChecked(True)
            devices_grid.addWidget(self.devices[i].gui_cb, i, 0)
            devices_grid.addWidget(label, i, 1)

        
        
        
        ### Create the "mode" radio box
        self.mode_box = QtGui.QGroupBox('Mode')
        self.mode_box.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed,
                                                      QtGui.QSizePolicy.Fixed))
        self.radio_realtime = QtGui.QRadioButton(self.mode_box)
        self.radio_realtime.setChecked(True) #default
        self.radio_playback = QtGui.QRadioButton(self.mode_box)
       
        realtime_label = QtGui.QLabel('Real-time')
        playback_label = QtGui.QLabel('Playback')
        
        radio_grid = QtGui.QGridLayout()
        radio_grid.addWidget(self.radio_realtime, 0, 0)
        radio_grid.addWidget(self.radio_playback, 1, 0)
        radio_grid.addWidget(realtime_label, 0, 1)
        radio_grid.addWidget(playback_label, 1, 1)
        self.mode_box.setLayout(radio_grid)
        
        
        
        ### Line
        self.line = QtGui.QFrame()
        self.line.setFrameShape(QtGui.QFrame.HLine)
        self.line.setFrameShadow(QtGui.QFrame.Sunken)
        
        ### Create the status box
        self.status_box = QtGui.QGroupBox('Status')
        self.status_box.setContentsMargins(15, 15, 15, 15)
        self.status_box.setSizePolicy(QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed,
                                                  QtGui.QSizePolicy.Fixed))
        status_layout = QtGui.QVBoxLayout()
        status_layout.setContentsMargins(15, 15, 15, 5)
        status_layout.setSpacing(25)
        self.status_edit = QtGui.QLineEdit(self.state)
        self.status_edit.setFixedSize(490, 26)
        self.status_edit.setReadOnly(True)
        
        self.launch_btn = QtGui.QPushButton('Launch nodes')
        self.launch_btn.clicked.connect(self.launch)
        
        status_layout.addWidget(self.status_edit)
        status_layout.addWidget(self.launch_btn, 0, QtCore.Qt.AlignHCenter)
        self.status_box.setLayout(status_layout)
        
        ### Line
        self.line2 = QtGui.QFrame()
        self.line2.setFrameShape(QtGui.QFrame.HLine)
        self.line2.setFrameShadow(QtGui.QFrame.Sunken)
        
        ### Line
        self.line3 = QtGui.QFrame()
        self.line3.setFrameShape(QtGui.QFrame.HLine)
        self.line3.setFrameShadow(QtGui.QFrame.Sunken)
        
    
        
    
        ### Create the main app layout
        main_layout = QtGui.QVBoxLayout()
        main_layout.setSizeConstraint(QtGui.QLayout.SetFixedSize)
        main_layout.setSpacing(10)
        main_layout.addWidget(self.devices_box)
        main_layout.addWidget(self.mode_box)
        main_layout.addWidget(self.line)
        main_layout.addWidget(self.status_box)
        main_layout.addWidget(self.line2)
        main_layout.addWidget(self.recorder.options_box)
        main_layout.addWidget(self.player.options_box)
        main_layout.addWidget(self.line3)
        main_layout.addWidget(self.recorder.media_box)
        main_layout.addWidget(self.player.media_box)
        main_layout.addWidget(QtGui.QLabel()) # extra space...
    
        self.setLayout(main_layout)
        self.setWindowTitle('Process Visualizer')
        self.setMinimumWidth(560)
        #self.setMinimumSize(560, 520)
        #self.setMaximumSize(560, 520)
        
        self.recorder.set_visible(False)
        self.player.set_visible(False)
        self.line2.setVisible(False)
        self.line3.setVisible(False)
        
    
    
    def launch(self):
        if self.state is DEFAULT:
            #launch nodes
            self.launcher = roslaunch.ROSLaunch()
            self.launcher.start()
            
            ### Update GUI etc..
            if self.radio_realtime.isChecked():
                self.state = REALTIME
                self.recorder.set_visible(True)
                enabled_topics = get_all_topics_to_record(self.devices)
                self.recorder.init(enabled_topics)
            else:
                self.state = PLAYBACK
                self.player.set_visible(True)
                self.player.init()
                
            self.devices_box.setDisabled(True)
            self.mode_box.setDisabled(True)
            self.line2.setVisible(True)
            self.line3.setVisible(True)
            self.launch_btn.setText('Quit all nodes')
            
            rospy.set_param('/process_visualizer/state', self.state)
            
            rviz_config = RVizConfig(self.pkg_path)
            
            ### Process all devices
            for device in self.devices:
                if device.gui_cb.isChecked():
                    
                    ### Type specific stuff...
                    if device.type == 'robot':
                        args = device.name + ' ' + self.config_path
                        node = roslaunch.core.Node('process_visualizer', 'robot_proc.py',
                                                   'robot_proc', namespace=device.name,
                                                   args=args, output='screen')
                        self.launcher.launch(node)
                        rviz_config.add(device.name, ['RobotModel', 'MarkerArray'])
                    elif device.type == 'depth_sensor':
                        args = device.name + ' ' + self.config_path
                        node = roslaunch.core.Node('process_visualizer', 'depth_proc.py',
                                                   'depth_proc', namespace=device.name,
                                                   args=args, output='screen')
                        self.launcher.launch(node)
                        rviz_config.add(device.name, ['PointCloud2'])
                    
                    ### Start the nodes
                    for node in device.nodes:
                        if node[1] == '' or node[1] == self.state:
                            self.launcher.launch(node[0])
                    
                    ### Start launch files
                    for launch in device.launches:
                        if launch[2] == '' or launch[2] == self.state:
                            args = ['roslaunch', launch[0], launch[1]]
                            launch_process = subprocess.Popen(args)
                            self.subprocesses.append(launch_process)
                    
                    ### Start static_transform_publishers
                    if device.base_frame != '':
                        args = device.xyz + ' ' + device.rpy + ' /base_link '
                        args = args + device.base_frame + ' 100'
                        node = roslaunch.core.Node('tf', 'static_transform_publisher',
                                    'base_frame_transformer',
                                    namespace=device.name,
                                    args=args, output='screen')
                        self.launcher.launch(node)
                    
                    ### Start static transformer for force frame
                    if device.force[0] != '':
		        args = device.force[2] + ' ' + device.force[3] + ' ' + device.force[1] + ' '
                        args += device.force[0] + ' 100'
                        node = roslaunch.core.Node('tf', 'static_transform_publisher',
                                    'force_frame_transformer',
                                    namespace=device.name,
                                    args=args, output='screen')
                        self.launcher.launch(node)

                    ### Add potential RViz displays
                    rviz_config.add(device.name, device.rviz)



            ### Start RViz
            rviz_config_path = rviz_config.close()
            #NEITHER OF THESE ATTEMPTS WORKS.... rviz crashes 30 ish % of the time :(
            #time.sleep(1)
            #cleaner_proc = subprocess.Popen(['rm -r ~/.rviz'], shell=True)
            #cleaner_proc.communicate()
            args = ['rosrun', 'rviz', 'rviz', '-d', rviz_config_path]
            rviz_proc = subprocess.Popen(args, stdin=subprocess.PIPE)
            self.subprocesses.append(rviz_proc)

            
        else:
            #quit
            self.stop_all()
            
            #self.kinect_proc_running = False
            self.state = DEFAULT
            
            self.devices_box.setEnabled(True)
            self.mode_box.setEnabled(True)
            self.recorder.set_visible(False)
            self.player.set_visible(False)
            self.line2.setVisible(False)
            self.line3.setVisible(False)
            self.launch_btn.setText('Launch nodes')



    def update_progress(self):
        #substate = ''
        if self.state is REALTIME:
            self.recorder.update_progress()
            substate = self.recorder.state
            if substate == 'recording':
                self.status_edit.setText('In real-time mode. Recording data')
            else:
                self.status_edit.setText('In real-time mode.')
        elif self.state is PLAYBACK:
            self.player.update_progress()
            self.status_edit.setText('In playback mode. Media is ' + self.player.state)
        else:
            self.status_edit.setText('Currently not doing anything...')
        
    
    
    def stop_all(self):
                
        if self.state is REALTIME:
            self.recorder.stop()
        elif self.state is PLAYBACK:
            self.player.stop()
        
        self.launcher.stop()
        
        #if self.state is not DEFAULT:
        #    self.rviz_proc.send_signal(subprocess.signal.SIGINT)
        for proc in self.subprocesses:
	    if proc is not None:
		proc.send_signal(subprocess.signal.SIGINT)
            
        print "Sent SIGINT"
        #time.sleep(5)
        for proc in self.subprocesses:
            print "PID:", proc.pid
            if proc is not None:
                proc.send_signal(subprocess.signal.SIGTERM)
        print "Sent SIGTERM"
        #time.sleep(2)
        for proc in self.subprocesses:
            if proc is not None:
                print proc.poll()
                proc.wait()
        self.subprocesses = []
        # Line below is really ugly, but was necessary to close all kinect drivers
        #subprocess.Popen(["/usr/bin/killall", "-9", "/usr/bin/XnSensorServer"])
        


# Makes sure the application catches SIGINT and exits cleanly
def main():
    
    rospy.init_node('process_visualizer')
    app = QtGui.QApplication(sys.argv)

    # Inspired by rxlaunch code
    signal.signal(signal.SIGINT, handle_sigint)
    timer = QtCore.QTimer()
    timer.start(250)
    timer.timeout.connect(lambda: None)  # Forces the interpreter to run every 250ms
    
    config_path = sys.argv[1]
    if not os.path.isfile(config_path):
        config_path = roslib.packages.get_pkg_dir('process_visualizer') + '/devices.xml'
        print "Could not find config file, or none specified. Using default: " + config_path
    
    form = LauncherApp(config_path)
    status_checker_timer = QtCore.QTimer()
    status_checker_timer.timeout.connect(form.update_progress)
    status_checker_timer.start(100)
    form.show()
    
    exit_code = -1
    try:
        exit_code = app.exec_()
    finally:
        form.stop_all()
    sys.exit(exit_code)



if __name__ == '__main__':
    main()



