#!/usr/bin/python
# -*- coding: utf-8 -*-
import sys
import os
import subprocess

import roslib
roslib.load_manifest('process_visualizer')
import rospy
import roslaunch

import xml.dom.minidom
import math
import numpy
from time import strftime

from std_msgs.msg import Float32MultiArray, Float32
from sensor_msgs.msg import JointState
from visualization_msgs.msg import MarkerArray, Marker
from geometry_msgs.msg import Quaternion, Point, Vector3

import tf
from tf.transformations import quaternion_from_euler#, quaternion_multiply
from xmlparser import XmlParser


# @todo: Change these values...
ARROW_MIN = 0.02
ARROW_MAX = 0.15
VEL_MIN = 0.05
VEL_MAX = 0.5
FORCE_MIN = 1
FORCE_MAX = 50



def arrow_length(len, min, max):
    if len <= min:
        return ARROW_MIN
    elif len >= max:
        return ARROW_MAX
    else:
        ans = (ARROW_MAX - ARROW_MIN) / (max - min) * (len - min) + ARROW_MIN
        return ans


def vectors_to_quaternion(vec1, vec2):
    
    # Transfer to numpy array format
    v1 = numpy.array((vec1.x, vec1.y, vec1.z))
    v2 = numpy.array((vec2.x, vec2.y, vec2.z))
    
    # Axis-Angle representation
    angle = math.acos(numpy.dot(v1, v2) / (numpy.linalg.norm(v1)*numpy.linalg.norm(v2)))
    axis = numpy.cross(v1, v2) / numpy.linalg.norm(numpy.cross(v1, v2))
    
    if (math.isnan(angle) or math.isnan(axis[0]) or math.isnan(axis[1]) or
            math.isnan(axis[2])):
        return numpy.array((0, 0, 0, 1))
    
    # Convert to quaternion
    sang = math.sin(angle*0.5)
    return numpy.array((axis[0]*sang, axis[1]*sang, axis[2]*sang, math.cos(angle*0.5)))



def get_param(name, value=None):
    private = "~%s" % name
    if rospy.has_param(private):
        return rospy.get_param(private)
    elif rospy.has_param(name):
        return rospy.get_param(name)
    else:
        return value



class RobotProcessor():
    def __init__(self, my_name, config_path):
        self.states_ready = False
        self.tcf_ready = False
        self.force_ready = False
        self.device = XmlParser(config_path).get_device(my_name)
        #self.tf_listener = tf.TransformListener()
        #self.tf_publisher = tf.TransformBroadcaster()
        self.force_length = 0
        self.force = Vector3(0, 0, 0)
        self.log_velocity_path = ( roslib.packages.get_pkg_dir('process_visualizer') +
                                   '/log/' + my_name + '_vel_' +
                                   strftime('%Y-%m-%d-%H-%M-%S') + '.txt' )
        self.log_force_path = ( roslib.packages.get_pkg_dir('process_visualizer') +
                                '/log/' + my_name + '_force_' +
                                strftime('%Y-%m-%d-%H-%M-%S') + '.txt' )
        
        if self.device is None:
            sys.exit('No device with the name ' + my_name + ' in the config file!')
        
        ### Get the URDF model
        modelpath = roslib.packages.get_pkg_dir(self.device.model[0])
        modelpath += '/' + self.device.model[1]
        urdf = ''
        if self.device.model[1][-6:] == '.xacro' and os.path.isfile(modelpath):
            args = []
            args.append('python')
            args.append(roslib.packages.get_pkg_dir('xacro') + '/xacro.py')
            args.append(modelpath)
            xacro_proc = subprocess.Popen(args, stdout=subprocess.PIPE)
            urdf = xacro_proc.communicate()[0]
        elif self.device.model[1][-5:] == '.urdf' and os.path.isfile(modelpath):
            urdf = open(modelpath, 'r').read()
        else:
            sys.exit('No valid model type defined for ' + my_name + 
                     '! Please add this as a <model> tag in the config file.')
        
        if urdf == '':
            sys.exit('ERROR: Could not read the model correctly.')
        

        ### Set parameter and start robot_state_publisher node
        rospy.set_param('/' + self.device.name + '/robot_description', urdf)
        robot_state_publisher_node = roslaunch.Node(package='robot_state_publisher',
                                                    node_type='state_publisher',
                                                    name='robot_state_publisher',
                                                    namespace=self.device.name)
        launcher = roslaunch.ROSLaunch()
        launcher.start()
        launcher.launch(robot_state_publisher_node)

        
        ### Fetch all the relevant topics
        self.states_topic = '' # joint states
        self.tcf_topic = '' # tool center frame
        self.force_topic = '' # force sensor data
        for top in self.device.topics:
            if top[1] == 'joint_states':
                self.states_topic = top[0]
            elif top[1] == 'tool_center_frame':
                self.tcf_topic = top[0]
            elif top[1] == 'force_sensor':
                self.force_topic = top[0]
                
        self.force_frame = self.device.force[0]
        
        if self.states_topic != '':
            self._init_states()
        else:
            #sys.exit('You need to specify a joint states topic in the xml configuration!')
            print 'IMPORTANT: You need to specify a joint states topic in the xml configuration!'
        
        if self.tcf_topic != '' or self.force_topic != '':
            self._init_markers()
        
        
    
    def _init_states(self):
        description = get_param('/' + self.device.name + '/robot_description')
        robot = xml.dom.minidom.parseString(description).getElementsByTagName('robot')[0]
        # IMPORTANT: joint_list must have the same order as the values on states_topic
        self.joint_list = []
        self.link_list = []
        
        # Find all non-fixed joints
        for child in robot.childNodes:
            if child.nodeType is child.TEXT_NODE:
                continue
            if child.localName == 'joint':
                jtype = child.getAttribute('type')
                if jtype == 'fixed':
                    continue
                name = child.getAttribute('name')
                self.joint_list.append(name)
                self.link_list.append(
                    (str(child.getElementsByTagName('parent')[0].getAttribute('link')),
                     str(child.getElementsByTagName('child')[0].getAttribute('link'))))

        rospy.Subscriber(self.states_topic, Float32MultiArray, self.states_cb, queue_size=1)
        self.states_pub = rospy.Publisher('/' + self.device.name + '/joint_states',
                                          JointState)
        
        # Set force frame to tool frame if it does not exist
        if self.force_frame == '':
	    self.force_frame = self.link_list[-1][1]
    
    
    
    def states_cb(self, msg):
        '''
        Callback function for the joint states
        '''
        #print '{0} - {1}'.format(msg.layout.dim[0].label, msg.data)
        self.js = JointState()
        self.js.header.stamp = rospy.Time.now()
        
        # Add data for all joints 
        for i in range(msg.layout.dim[0].size):
            self.js.name.append(str(self.joint_list[i]))
            self.js.position.append(msg.data[i])
        
        # Add the tool joint
        # @todo: improve this??
        self.js.name.append(str(self.joint_list[msg.layout.dim[0].size]))
        self.js.position.append(0)
        
        self.states_ready = True
        self.publish_states()
        #self.sync()
        
    
    
    
    
    def _init_markers(self):
        '''
        Initiates the markerarray and tcf/force topics
        '''
        self.rate = 1.0
        self.vel_length = 0
        self.velocity = Vector3(0.0, 0.0, 0.0)
        self.marker_pub = rospy.Publisher('/' + self.device.name + '/marker_array',
                                          MarkerArray)
        self.marker_array = MarkerArray()
        
        
        
        
        if self.tcf_topic != '':
            rospy.Subscriber(self.tcf_topic, Float32MultiArray, self.tcf_cb, queue_size=1)
            
            # Set initial values used to calculate the velocity
            self.prev_pos = Point(0.0, 0.0, 0.0)
            self.curr_pos = Point(0.0, 0.0, 0.0)
            self.prev_time = rospy.Time()
            self.curr_time = rospy.Time()
            
            # Initialize tool velocity vector marker
            self.tool_vel = Marker()
            self.tool_vel.header.frame_id = self.device.base_frame
            self.tool_vel.ns = "tool_vel"
            self.tool_vel.id = 0
            self.tool_vel.type = Marker.ARROW
            self.tool_vel.action = Marker.ADD
            self.tool_vel.scale.x = 0.2
            self.tool_vel.scale.y = 0.2
            self.tool_vel.scale.z = 0.1 #length
            self.tool_vel.color.a = 1.0
            self.tool_vel.color.r = 0.0
            self.tool_vel.color.g = 1.0
            self.tool_vel.color.b = 0.0
            self.marker_array.markers.append(self.tool_vel)
            
            self.log_velocity = open(self.log_velocity_path, 'w')
        
        
        if self.force_topic != '':
            rospy.Subscriber(self.force_topic, Float32MultiArray, self.force_cb, queue_size=1)
            
            # Initialize the force vector marker
            self.tool_force = Marker()
            #self.tool_force.header.frame_id = '/' + self.device.name + '/tool' # @todo: FIX
            #self.tool_force.header.frame_id = self.device.base_frame
            #if self.tcf_topic != '':
            #    self.tool_force.header.frame_id = self.device.base_frame
            #else:
            #    self.tool_force.header.frame_id = str(self.link_list[-1][1])
            #self.tool_force.header.frame_id = str(self.link_list[-1][1])
            self.tool_force.header.frame_id = str(self.force_frame)
            self.tool_force.ns = 'tool_force'
            self.tool_force.id = 1
            self.tool_force.type = Marker.ARROW
            self.tool_force.action = Marker.ADD
            self.tool_force.scale.x = 0.2
            self.tool_force.scale.y = 0.2
            self.tool_force.scale.z = 0.1 #length
            self.tool_force.pose.position.x = 0.0
            self.tool_force.pose.position.y = 0.0
            self.tool_force.pose.position.z = 0.0
            self.tool_force.color.a = 1.0
            self.tool_force.color.r = 1.0
            self.tool_force.color.g = 0.0
            self.tool_force.color.b = 0.0
            self.marker_array.markers.append(self.tool_force)
            
            self.log_force = open(self.log_force_path, 'w')
        
        
        # Initialize the text description marker
        self.text_desc = Marker()
        self.text_desc.header.frame_id = self.device.base_frame
        self.text_desc.ns = "text_desc"
        self.text_desc.id = 2
        self.text_desc.type = Marker.TEXT_VIEW_FACING
        self.text_desc.action = Marker.ADD
        self.text_desc.scale.z = 0.04 #text size
        self.text_desc.pose.position.x = 0.5
        self.text_desc.pose.position.y = 0.5
        self.text_desc.pose.position.z = 0.0
        self.text_desc.color.a = 1.0
        self.text_desc.color.r = 1.0
        self.text_desc.color.g = 1.0
        self.text_desc.color.b = 1.0
        self.marker_array.markers.append(self.text_desc)
        
    

    def tcf_cb(self, msg):
        '''
        Callback function for /tcf (tool center frame) topic
        Calculates the velocity and calls sync() function
        '''
        
        # Update "current" values
        self.curr_pos.x = msg.data[0]
        self.curr_pos.y = msg.data[1]
        self.curr_pos.z = msg.data[2]
        self.curr_time = rospy.Time.now()
        
        
        if self.curr_time.to_sec() - self.prev_time.to_sec() > 1:
            self.rate = float(get_param('/process_visualizer/playback_rate', '1'))
        
        
        # Check if this is the first time
        if self.prev_time.to_sec() == 0.0:
            self.prev_pos.x = self.curr_pos.x
            self.prev_pos.y = self.curr_pos.y
            self.prev_pos.z = self.curr_pos.z
            self.prev_time = self.curr_time
            return
        
        # @todo: ATM delta_t is wrong if the program is stepping in playback mode
        delta_t = (self.curr_time.to_sec()-self.prev_time.to_sec()) * self.rate
        self.velocity = Vector3((self.curr_pos.x-self.prev_pos.x) / delta_t,
                           (self.curr_pos.y-self.prev_pos.y) / delta_t,
                           (self.curr_pos.z-self.prev_pos.z) / delta_t)
        
        self.vel_length = math.sqrt(self.velocity.x**2 + self.velocity.y**2 +
                                    self.velocity.z**2)
        
        quat = vectors_to_quaternion(Vector3(1.0, 0.0, 0.0), self.velocity)
        
        # Update the tool_vel marker
        self.tool_vel.header.stamp = self.curr_time
        self.tool_vel.scale.z = arrow_length(self.vel_length, VEL_MIN, VEL_MAX)
        self.tool_vel.pose.position = Point(msg.data[0], msg.data[1], msg.data[2])
        self.tool_vel.pose.orientation = Quaternion(*quat)
        
        # Reset variables
        self.prev_pos.x = self.curr_pos.x
        self.prev_pos.y = self.curr_pos.y
        self.prev_pos.z = self.curr_pos.z
        self.prev_time = self.curr_time
        
        self.tcf_ready = True
        self.publish_markers()
        #self.sync()
        
        # Log to file (velocity + position...)
        self.log_velocity.write(str(self.curr_time) + ' ' + str(self.velocity.x) + ' ' +
                                str(self.velocity.y) + ' ' + str(self.velocity.z) + ' ' +
                                str(self.curr_pos.x) + ' ' +
                                str(self.curr_pos.y) + ' ' + str(self.curr_pos.z) + '\n')
        self.log_velocity.flush()

    def force_cb(self, msg):
        '''
        Callback function for force (/torque) topic
        '''
        
        self.force = Vector3(msg.data[0], msg.data[1], msg.data[2])
        self.force_length = math.sqrt(self.force.x**2 + self.force.y**2 + self.force.z**2)
        quat = vectors_to_quaternion(Vector3(1.0, 0.0, 0.0), self.force)
        
        #if self.tcf_topic != '':
        #    self.tool_force.pose.position = self.tool_vel.pose.position
        #    try:
        #        quat2 = self.tf_listener.lookupTransform(self.device.base_frame,
        #                                         str(self.link_list[-1][1]),
        #                                         rospy.Time(0))[1]
        #    except:
        #        quat2 = numpy.array((0.0, 0.0, 0.0, 1.0))
        #s    quat = quaternion_multiply(quat, quat2)
        
        # Update the tool_force marker
        self.tool_force.header.stamp = rospy.Time.now()
        self.tool_force.scale.z = arrow_length(self.force_length, FORCE_MIN, FORCE_MAX)
        self.tool_force.pose.orientation = Quaternion(*quat)
        
        self.force_ready = True
        self.publish_markers()
        #self.sync()
    
        # Log to file
        self.log_force.write(str(rospy.Time.now()) + ' ' + str(self.force.x) + ' ' +
                                str(self.force.y) + ' ' + str(self.force.z) + '\n')
        self.log_force.flush()

    
    def sync(self):
        '''
        Sync messages, to publish everything at the same time
        '''
        #self.publish()
        #if self.device.name == 'ursec_upper':
        #    print (str(self.states_ready) + ' ' +
        #           str(self.tcf_ready or self.tcf_topic == '') + ' ' +
        #           str(self.force_ready or self.force_topic == '') + ' ' +
        #           str((self.tcf_topic == '' or self.force_topic == '' or 
        #               self.tcf_ready and self.force_ready)))
        #if (self.states_ready and
        #        (self.tcf_topic == '' or self.force_topic == '' or 
        #         self.tcf_ready and self.force_ready)):
        #    self.publish()
        #if self.states_ready and self.tcf_ready and self.force_ready:
        #    self.publish()
        
        #if (self.states_ready and
        #        (self.tcf_ready or self.tcf_topic == '') and
        #        (self.force_ready or self.force_topic == '')):
        #    self.publish()
    
    
    def publish_markers(self):
        # Update the text_desc marker
        if self.tcf_topic != '' or self.force_topic != '':
            self.text_desc.header.stamp = rospy.Time.now()
            self.text_desc.text = '*** {0} ***'.format(self.device.name)
            if self.tcf_topic != '':
                if (math.isnan(self.vel_length) or 
                        math.isinf(self.vel_length) or
                        math.fabs(self.vel_length) < 0.001):
                    self.tool_vel.action = Marker.DELETE
                else:
                    self.tool_vel.action = Marker.ADD
                    self.text_desc.text += '\nTool velocity: {0}m/s ({1}, {2}, {3})'.format(
                            round(self.vel_length, 3), round(self.velocity.x, 3),
                            round(self.velocity.y, 3), round(self.velocity.z, 3))
            if self.force_topic != '':
                self.text_desc.text += '\nTool force: {0}N ({1}, {2}, {3})'.format(
                            round(self.force_length, 3), round(self.force.x, 3),
                            round(self.force.y, 3), round(self.force.z, 3))
                # Use tcf info instead, if this is available
                #if self.tcf_topic != '':
                    
            #try:
            #    (ee_pos, ee_rot) = self.tf_listener.lookupTransform(self.device.base_frame,
            #                                                    str(self.link_list[-1][1]),
            #                                                    rospy.Time(0))
            #    self.tool_force.pose.position = Point(*ee_pos)
            #except:
            #    self.tool_force.pose.position = Point(0.0, 0.0, 0.0)
            #print (ee_pos, ee_rot)
            #
            
            self.marker_pub.publish(self.marker_array)
    
        
    def publish_states(self):
        '''
        Updates the text_desc marker, and publishes everything
        '''
        
        self.states_pub.publish(self.js)
        
        

        # Reset variables
        #self.states_ready = False
        #self.tcf_ready = False
        #self.force_ready = False

        



if __name__ == '__main__':
    try:
        rospy.init_node('robot_proc')
        
        '''
        argv[1]: Robot name
        argv[2]: Path to XML configuration file
        '''
        RobotProcessor(sys.argv[1], sys.argv[2])
        
        rospy.spin()
    except rospy.ROSInterruptException: pass

