# -*- coding: utf-8 -*-
import sys
import os
from xml.dom.minidom import parse
from roslaunch.core import Node
import roslib



class Device(object):
    def __init__(self, name, type, model=None, topics=[],
                 nodes=[], launches=[], base_frame='',
                 xyz='0 0 0', rpy='0 0 0', force=('','','0 0 0','0 0 0'), rviz=[]):
        self.name = name
        self.type = type
        self.model = model
        self.topics = topics
        self.nodes = nodes
        self.launches = launches
        self.base_frame = base_frame
        self.xyz = xyz
        self.rpy = rpy
        self.force = force
        #self.xyz = tuple([float(i) for i in xyz.split(' ')])
        #self.rpy = tuple([float(i) for i in rpy.split(' ')])
        self.rviz = rviz
        self.gui_cb = None
    
    def __str__(self):
        return self.name + " " + self.type + " " + self.model


class XmlParser():
    '''
    Class used to parse devices.xml files
    '''
    
    def __init__(self, path):
        #self.file = path + '/devices.xml'
        
        root = parse(path).getElementsByTagName('visualizer')[0]
        self.devices = root.getElementsByTagName('device')
        
        # @todo: CHECK IF THE FILE IS VALID

    
    
    def get_device(self, name):
        for elem in self.devices:
            if str(elem.getAttribute('name')) == name:
                return element_to_device(elem)
        return None
    
    
    def get_all(self):
        '''
        @return: a list of 'Device' objects
        '''
        devs = []
        for elem in self.devices:
            devs.append(element_to_device(elem))
        return devs


def element_to_device(elem):
    name = str(elem.getAttribute('name'))
    type = str(elem.getAttribute('type'))
    model = _get_model(elem)
    topics = _get_topics(elem)
    nodes = _get_nodes(elem)
    launches = _get_launches(elem)
    base_frame = ''
    xyz = '0 0 0'
    rpy = '0 0 0'
    for cur_base in elem.getElementsByTagName('base'):
        base_frame = str(cur_base.getAttribute('frame'))
        xyz = str(cur_base.getAttribute('xyz'))
        rpy = str(cur_base.getAttribute('rpy'))
    if xyz == '':
        xyz = '0 0 0'
    if rpy == '':
        rpy = '0 0 0'
    force = _get_force(elem)
    rviz = _get_rviz(elem)
    device = Device(name, type, model, topics, nodes, launches,
                    base_frame, xyz, rpy, force, rviz)
    return device

def _get_model(device):
    '''
    @return: (pkg, path)
    '''
    for elem in device.getElementsByTagName('model'):
        return ( str(elem.getAttribute('pkg')), str(elem.getAttribute('path')) )
    return ('', '')

def _get_topics(device):
    '''
    @type device: element
    
    @return: list of tuples: (name, type, record) [str, str, bool]
    '''
    topics = []
    for elem in device.getElementsByTagName('topic'):
        if elem.hasAttribute('ns'):
            ns = '/' + str(elem.getAttribute('ns'))
        else:
            ns = ''
        if str(elem.getAttribute('record')).lower() == 'false':
            record = False
        else:
            record = True
        topics.append( (ns + '/' + elem.getAttribute('name'),
                        elem.getAttribute('type'),
                        record) )
    return topics

def _get_nodes(device):
    '''
    @param device: the dom device
    @type device: element
    
    @return: list of tuples (roslaunch.Node, ifstate)
    
    Comment: Could/should use some of roslaunch.xmlloader / .core
             functionality instead?
    '''
    nodes = []
    for elem in device.getElementsByTagName('node'):
        pkg = str(elem.getAttribute('pkg'))
        type = str(elem.getAttribute('type'))
        name = str(elem.getAttribute('name'))
        if elem.hasAttribute('ns'):
            ns = '/' + str(elem.getAttribute('ns'))
        else:
            ns = '/' + str(device.getAttribute('name'))
        args = str(elem.getAttribute('args'))
        output = str(elem.getAttribute('output'))
        if output == '':
            output = 'screen'
        ifstate = str(elem.getAttribute('ifstate')).lower()
        
        node = Node(pkg, type, name,
                    namespace=ns,
                    args=args,
                    output=output)
        nodes.append( (node, ifstate) )
    return nodes

def _get_launches(device):
    '''
    @return: list of tuples: (pkg, type, ifstate)
    '''
    launches = []
    for elem in device.getElementsByTagName('launch'):
        ifstate = str(elem.getAttribute('ifstate')).lower()
        pkg = str(elem.getAttribute('pkg'))
        type = str(elem.getAttribute('type'))
        
        launches.append( (pkg, type, ifstate) )
    return launches

def _get_rviz(device):
    '''
    @type device: element
    '''
    displays = []
    for elem in device.getElementsByTagName('rviz'):
        type = str(elem.getAttribute('type'))
        displays.append(type)
    return displays

def _get_force(device):
    for force in device.getElementsByTagName('force'):
        parent = str(force.getAttribute('parentframe'))
	frame = str(force.getAttribute('frame'))
	xyz = str(force.getAttribute('xyz'))
	rpy = str(force.getAttribute('rpy'))
	return (frame, parent, xyz, rpy)
    return ('', '', '0 0 0', '0 0 0')


