#!/usr/bin/python
# -*- coding: utf-8 -*-

import roslib
roslib.load_manifest('sewing_cell')
import rospy

from std_msgs.msg import Float32MultiArray, UInt16, Float32
from visualization_msgs.msg import MarkerArray, Marker
from geometry_msgs.msg import Quaternion, Point#, Vector3

import math
import tf
from tf.transformations import quaternion_from_euler

from robot_proc import ARROW_MAX#, arrow_length

LINE_VALUE_MIN = 5
LINE_VALUE_MAX = 505
def edge_arrow_length(curr, setp):
    if curr < LINE_VALUE_MIN or curr > LINE_VALUE_MAX:
        return 0
    else:
        return float(curr - setp) / float(LINE_VALUE_MAX - setp) * ARROW_MAX


class SewingTransformer():
    def __init__(self):
        self.name = 'sewing_machine'
        self.base_frame = '/' + self.name + '/base'
        self.needle_frame = '/' + self.name + '/needle'
        self.edge_upper_topic = '/edge_finder_upper'
        self.edge_lower_topic = '/edge_finder_lower'
        self.feed_rate_topic = '/feed_rate_est'
        
        self.markers_pub = rospy.Publisher('/' + self.name + '/marker_array', MarkerArray)
        self.marker_array = MarkerArray()
        
        
        # Initialize tool velocity vector marker
        self.sewing_machine = Marker()
        self.sewing_machine.header.frame_id = self.base_frame
        self.sewing_machine.ns = self.name
        self.sewing_machine.id = 0
        self.sewing_machine.type = Marker.MESH_RESOURCE
        self.sewing_machine.mesh_resource = "package://sewing_cell/meshes/sewing_machine.dae"
        self.sewing_machine.mesh_use_embedded_materials = True
        self.sewing_machine.action = Marker.ADD
        self.sewing_machine.scale.x = 1.0
        self.sewing_machine.scale.y = 1.0
        self.sewing_machine.scale.z = 1.0
        self.marker_array.markers.append(self.sewing_machine)
        
        
        # Initialize upper edge sensor vector marker
        self.edge_upper_setpoint = rospy.get_param('/sewing_machine/setpoint_upper', 255)
        rospy.Subscriber(self.edge_upper_topic, UInt16, self.edge_upper_cb, queue_size=1)
        
        self.edge_upper = Marker()
        self.edge_upper.header.stamp = rospy.Time.now()
        self.edge_upper.header.frame_id = self.needle_frame
        self.edge_upper.ns = "edge_upper"
        self.edge_upper.id = 0
        self.edge_upper.type = Marker.ARROW
        self.edge_upper.action = Marker.ADD
        self.edge_upper.scale.x = 0.15
        self.edge_upper.scale.y = 0.15
        self.edge_upper.scale.z = 0.0 #length
        self.edge_upper.color.a = 1.0
        self.edge_upper.color.r = 0.0
        self.edge_upper.color.g = 1.0
        self.edge_upper.color.b = 1.0
        self.edge_upper.pose.position = Point(0.0, 0.0, 0.02)
        quat = quaternion_from_euler(0, 0, math.pi/2)
        self.edge_upper.pose.orientation = Quaternion(*quat)
        self.marker_array.markers.append(self.edge_upper)
        
        
        # Initialize lower edge sensor vector marker
        self.edge_lower_setpoint = rospy.get_param('/sewing_machine/setpoint_lower', 255)
        rospy.Subscriber(self.edge_lower_topic, UInt16, self.edge_lower_cb, queue_size=1)
        
        self.edge_lower = Marker()
        self.edge_lower.header.stamp = rospy.Time.now()
        self.edge_lower.header.frame_id = self.needle_frame
        self.edge_lower.ns = "edge_lower"
        self.edge_lower.id = 1
        self.edge_lower.type = Marker.ARROW
        self.edge_lower.action = Marker.ADD
        self.edge_lower.scale.x = 0.15
        self.edge_lower.scale.y = 0.15
        self.edge_lower.scale.z = 0.0 #length
        self.edge_lower.color.a = 1.0
        self.edge_lower.color.r = 1.0
        self.edge_lower.color.g = 1.0
        self.edge_lower.color.b = 0.0
        self.edge_lower.pose.position = Point(0.0, 0.0, 0.0)
        quat = quaternion_from_euler(0, 0, math.pi/2)
        self.edge_lower.pose.orientation = Quaternion(*quat)
        self.marker_array.markers.append(self.edge_lower)
    
        
        # Initialize feed rate vector marker
        rospy.Subscriber(self.feed_rate_topic, Float32, self.feed_rate_cb, queue_size=1)
        
        self.feed_rate = Marker()
        self.feed_rate.header.stamp = rospy.Time.now()
        self.feed_rate.header.frame_id = self.needle_frame
        self.feed_rate.ns = "feed_rate"
        self.feed_rate.id = 2
        self.feed_rate.type = Marker.ARROW
        self.feed_rate.action = Marker.ADD
        self.feed_rate.scale.x = 0.15
        self.feed_rate.scale.y = 0.15
        self.feed_rate.scale.z = 0.0 #length
        self.feed_rate.color.a = 1.0
        self.feed_rate.color.r = 1.0
        self.feed_rate.color.g = 0.0
        self.feed_rate.color.b = 1.0
        #self.feed_rate.pose.position = Point(0.0, 0.0, 0.0)
        quat = quaternion_from_euler(0, 0, math.pi)
        self.feed_rate.pose.orientation = Quaternion(*quat)
        self.marker_array.markers.append(self.feed_rate)
    
        
        
        # Initialize the text description marker
        self.text_desc = Marker()
        self.text_desc.header.frame_id = "/base_link"
        self.text_desc.ns = "text_desc"
        self.text_desc.id = 3
        self.text_desc.type = Marker.TEXT_VIEW_FACING
        self.text_desc.action = Marker.ADD
        self.text_desc.scale.z = 0.04 #text size
        self.text_desc.pose.position.x = 0.1
        self.text_desc.pose.position.y = -1.0
        self.text_desc.pose.position.z = 0.0
        self.text_desc.color.a = 1.0
        self.text_desc.color.r = 1.0
        self.text_desc.color.g = 1.0
        self.text_desc.color.b = 1.0
        self.marker_array.markers.append(self.text_desc)
        


    def edge_upper_cb(self, msg):
        self.edge_upper_previous = msg.data
        self.edge_upper.scale.z = edge_arrow_length(msg.data, self.edge_upper_setpoint)
        #self.publish_markers()
        
    def edge_lower_cb(self, msg):
        self.edge_lower_previous = msg.data
        self.edge_lower.scale.z = edge_arrow_length(msg.data, self.edge_lower_setpoint)
        #self.publish_markers()

    def feed_rate_cb(self, msg):
        self.feed_rate_previous = msg.data
        self.feed_rate.scale.z = 2*msg.data # perhaps scale this a bit?
    
    def publish_markers(self):
        self.sewing_machine.header.stamp = rospy.Time.now()
        self.edge_upper.header.stamp = rospy.Time.now()
        self.edge_lower.header.stamp = rospy.Time.now()
        self.feed_rate.header.stamp = rospy.Time.now()
        self.text_desc.header.stamp = rospy.Time.now()
        
        self.text_desc.text = '*** Sewing Machine ***'
        
        if math.fabs(self.edge_upper.scale.z) < 0.001:
            self.edge_upper.action = Marker.DELETE
        else:
            self.edge_upper.action = Marker.ADD
            self.text_desc.text += '\nLine sensor upper: {0}'.format(
                                   self.edge_upper_previous)
            
        if math.fabs(self.edge_lower.scale.z) < 0.001:
            self.edge_lower.action = Marker.DELETE
        else:
            self.edge_lower.action = Marker.ADD
            self.text_desc.text += '\nLine sensor lower: {0}'.format(
                                   self.edge_lower_previous)
        
        if math.fabs(self.feed_rate.scale.z) < 0.001:
            self.feed_rate.action = Marker.DELETE
        else:
            self.feed_rate.action = Marker.ADD
            self.text_desc.text += '\nFeed rate: {0}'.format(
                                   self.feed_rate_previous)
        
        self.markers_pub.publish(self.marker_array)

    

def main():
    rospy.init_node('sewing_machine')
    
    st = SewingTransformer()
    #st.publish_marker()
    #rospy.spin()
    r = rospy.Rate(10)
    while not rospy.is_shutdown():
        st.publish_markers()
        r.sleep()
    


if __name__ == '__main__':
    main()
